#pragma once
#include "device.h"
#include "PluginHeader.h"

#include "ScopedPacketAccess.h"

/** Parent class for devices that create data
///	@ingroup devices

*/
class Generator : public Device  
{
public:
	Generator();
	virtual ~Generator();

	virtual int GetDeviceFamily() { return TYPE_GENERATOR; };
	virtual dword Color() { return (0x068b12c); }; 

	virtual bool isMaskable() { return true; };
	virtual bool hasOrigin() { return true; };
	virtual bool isExplorable() { return getOriginType() != 1; };	// is now Explorer-stable in Pro mode
	virtual bool hasLocalOverride() { return true; };
protected:
	// default generator behavior is to mask against zero on all outputs.
	virtual bool doMaskCreateNewCopies(std::vector<PacketPointer> &results,  BuildContext &context);
};

class OutputData;

/** Parent class of Devices that export data.
	Creating an output that is not derived from this class can lead to undefined behavior; 
	World Machine will not consider your device an Output unless its parent is the Output class.

///	@ingroup devices
*/
class  Output : public Device  
{
public:
	Output();
	virtual ~Output();

	virtual void doClone(Device *other);			// Copy data from other to this

	enum OUTPUT_MODE { OUTPUT_NORMAL, OUTPUT_TILED};

	virtual int GetDeviceFamily() { return TYPE_OUTPUT; };
	virtual dword Color() { return (0x03030cc); };

	/** Given the set of output data, export according to the class of Output device.
	@param outset		All inputs the output needs to produce an output. Typically (but does not need to) holds the current final-build contents of the device
	@param output_mode	Normal or Tiled build; this controls how the Output handles file naming.
	*/
	virtual bool doOutput(OutputData &outset, OUTPUT_MODE output_mode) = 0;

	virtual bool RunUI(); // Output::RunUI() always returns false; (because) its impossible to invalidate data contents since we dont operate on them

	virtual ACTIVATE_RESULT ActivateOnNeed(BuildContext &bi);

	/// Output provides a default specialization for Activate. Unless you need to do something unique on Activate, you can simply implement your doOutput call and be good
	/// to go! If you do need to perform your own activate, make sure you call Output::Activate() before performing your own actions. More details available
	virtual bool Activate(BuildContext &context);
	
	virtual bool allowTileBlend() { return true; };

	virtual bool isExporting(OUTPUT_MODE mode) = 0;	///< Return false if output is view-only or is otherwise not currently exporting (this helps tiled build and others save time)
	virtual bool isMaskable() { return false; };

	virtual bool RunAuxAction(int index);		// Run extra actions defined by this Device
	virtual int  getAuxActionSize();			// The number of actions defined
	virtual char *getAuxActionName(int index);	// The name of the extra action

	/// Convenience wrapper to World retrieve output information
	OutputData RetrieveOutputData();
protected:
	bool always_export;
	bool participate_tiled;

	/// Override this to change how the preview packet is created from the output inputs
	/// default is simply the first input.
	virtual PacketPointer doMakePreviewPacket(const OutputData &outset);

	/// This utility function decorates the filename your device is exporting.
	/// You can either pass a pointer to the parameter containing the filename (in which case it will retrieve the filename string)
	/// or you can pass the raw filename directly. The returned name is decorated appropriately, for example:
	/// testoutput.ter -> testoutput_tile_x1_y2.ter
	WMString getExportFilename(const Parameter *filename);
	WMString getExportFilename(const char *filename);

	void StoreOutputView(BuildContext &context, PacketPointer preview_packet);
private:
};

class DeviceResults;

/** Encapsulation of the set of inputs piped to an output device
	@ingroup devices
	@ingroup data	
*/
class OutputData {
public:
	OutputData() {};
	OutputData(DeviceResults &context);
	OutputData(Device &dev);

	bool valid() const { if (inputs.size() > 0) return inputs[0].get() != NULL; return false; };

	int num() const { return (int) inputs.size(); };
	PacketPointer get(int i) const { if (i < (int) inputs.size()) { return inputs[i].get(); } return NULL; };

	void set(int i, PacketPointer const& data) { assert(i < (int) inputs.size()); inputs[i] = ScopedPacketAccess(data); };
	// Remove and deallocate contents
	void cleanup() { inputs.clear(); };
private:
	std::vector<ScopedPacketAccess> inputs;
};
