#pragma once
#include "Globals.h"

namespace WorldMachine {

/**
	Interface class for running commands on the World Machine UI thread using WorldMachine::core->RunFunction().

	Although you can use this directly, consider using the wrapper template and WorldMachine::RunFunction; they provide the same functionality
	with a drastic reduction of boilerplate and no further loss of efficiency.
*/
class RunFunctionInterface {
public:
	/// Implement whatever you need to have run in your Run() function.
	virtual void Run() = 0;

	/// You must implement this function to destroy the message on your side; the WM core will call Release() after executing your function.
	virtual void Release() = 0;
};


/** Wrapper template around the RunFunctionInterface to simplify the boilerplate of having to create new interface implementations
	just to send a command to the main thread.

	For more details, see the documentation for the RunFunction() free function!
*/
template < class YourFunc >
class RunFuncWrapper : public RunFunctionInterface {
public:
	virtual void Run() override {
		func();
	};

	virtual void Release() override {
		delete this;
	}
private:
	RunFuncWrapper(YourFunc f) : func(f) {};
	RunFuncWrapper(RunFuncWrapper const& other);
	YourFunc func;

	template <class OtherFunc> friend void RunFunction(Device *, OtherFunc);
};

/** Convenience function to easily have some operation performed from the main UI thread. You can avoid all boilerplate requirements
	that implementing the RunFunctionInterface requires and simply call this function.
	Supply a lambda, function object, or function pointer. ie:

		@code
		WorldMachine::RunFunction(currentdevice, []() {
			// Do something in this lambda that will be posted to run on the main thread
		});
		@endcode

		Compare the above to the sheer amount of interface-writing boilerplate that would be required if writing
		directly to the interface:

		@code
		class MyFuncInt : public WorldMachine::RunFunctionInterface {
			Device *dev;
		public:
			MyFunc(Device *d) : dev(d) {};

			virtual void Run() override {
				// Do something from the main thread!
			};

			virtual void Release() override {
				delete this;
			};
		};
		// Launch our interface implementation
		WorldMachine::core->app->RunFunction(currentdevice, new MyFuncInt(currentdevice));
		@endcode
*/
template < class CreateType >
void RunFunction(Device *dev, CreateType f) { 
	WorldMachine::core->app->RunFunction(dev, new RunFuncWrapper<CreateType>(f)); 
};


}
