#pragma once

#include <vector>
#include "parameter.h"

/// A unique identifier for a Device and Preset combo.
struct PresetName {
	char tag[4];				// tag of device owner
	std::string presetname;		// filename of preset
};

/// Encapsulation of a set of parameters for the PresetManager
struct PresetData {
	char tag[4];				// tag of device owner
	ParameterSet params;
};

/** Encapsulates the functionality of presets for devices.
	@ingroup devices
 It does NOT store the actual data of the preset.. however it manages scan/load/store of them
 Used by GenericDlg, and YOUR dialog that wants to use device presets

@par How to Use:

	-# SetPresetBase to initialize the preset manager for your device
	-# ScanPresets() will populate the manager with any presets existing on disk
	-# Iterate through the populated presets with numPresets(), getPreset().
	   Note that adding a new preset with addPreset() only adds it to the LIST, not to the actual set of presets on disk
	-# Load or save your presets as needed with LoadPreset() or SavePreset()


*/
class PresetManager
{
public:
	PresetManager(void);
	~PresetManager(void);

	/** Initialize manager for a given device.
		Given an owner device and tag ( usually Device::GetStaticTag() ), and the preset folder of World Machine (usually WorldMachine::core->app->preset_path)
	*/
	void SetPresetBase(Device *_owner, const char *tag, const char *WMdir);

	bool ScanPresets();												///< create list of presets for a given tag
	bool LoadPreset(const PresetName &name, PresetData &preset);	///< load a preset from disk
	bool SavePreset(const PresetName &name, PresetData &preset);	///< save a preset to disk
	bool DelPreset(const PresetName &name);							///< remove a preset from disk

	int			numPresets()				{ return (int) presetlist.size(); };
	PresetName &getPreset(int index)		{ assert(index >= 0); assert(index < (int) presetlist.size()); return presetlist[index]; };
	void		addPreset(const char *str);
protected:
	bool ready;							// has basedir been set?
	Device *owner;
	char tag_id[4];						// Tag ID for the current list of presets 
	std::string basedir;
	std::vector<PresetName> presetlist;
};
