#pragma once

class HField;
class BMPacket;
class PacketContext;

#include "common.h"
#include "Context.h"
#include "GridPacket.h"
#include "SimpleRand.h"
#include "color.h"
#include "RescalerInfo.h"

namespace WorldMachine {
namespace Rescale {
	/* *** HIGH LEVEL PRIMITIVES *** */
	/// Highest level, polymorphic rescaling.
	/// Rescale the given packet to a new size and return it.
	/// If the packet is not of a supported (grid) type, will return NULL
	PacketPointer RescalePacket(PacketPointer wmp, Coord dims, RescaleInfo type = RescaleInfo(DEFAULT));
	PacketPointer RescalePacket(PacketPointer wmp, PacketContext &destcontext, RescaleInfo type = RescaleInfo(DEFAULT));

	/// Additional type-specific convenience functions that do the casting for you...
	HFPointer RescaleHF(HFPointer const&hf, Coord dims, RescaleInfo type = RescaleInfo(DEFAULT));
	BMPointer RescaleBMP(BMPointer const&hf, Coord dims, RescaleInfo type = RescaleInfo(DEFAULT));

	/// Utility functions to adjust a resolution pair according to a hard limit; the returned resolution will be aspect-adjusted to fit the provided res.
	Coord LimitSize(const Coord &size, int limit);
	Coord LimitSize(const Coord &size, const Coord &limit); // optionally make sure we fit to a rectangle-sized limit

	/* *** MID LEVEL PRIMITIVES *** */
	/// RescaleExisting resamples the source packet into the destination one, using the method specified.
	/// you can either provide the first template parameter to specify a resampling type, such as:
	///  RescaleExisting<Rescale::BILINEAR>( src, dest);
	/// or provide a full RescaleInfo parameter containing additional rescaling data, such as:
	///  RescaleExisting(src, dest, rescale_info);
	template< RESCALE_TYPE Method, typename T >
	void RescaleExisting( WMGridPacket<T> const&src, WMGridPacket<T> &dest, RescaleInfo method = RescaleInfo(Method) );

	/*** LOW LEVEL PRIMITIVES ***/
	/// The resamplig kernel primitive takes a 4x4 data area, with x1,y1 defined as the floor of the point to evaluate.
	/// The Nearest and Bilinear primitves require only x1,y1 through x2,y2 to be filled.
	/// The primitive version is useful when you have more exotic re-sampling needs than just turning one heightfield into another.
	/// you could use these for sampling from a rotational/etc kernels, for example.
	template <typename T>
	T Resample_Kernel(T src[4][4], CoordF uv, RescaleInfo type = RescaleInfo(DEFAULT), FractalRescaleInfo fractalinfo = FractalRescaleInfo());

// C++11 extern template define the grid packet rescalings that the PDK supports.
// You can currently only use the Resample_Kernel with a float or color datatype. Other types are not provided publicly.
#ifndef RESCALER_INSTANTIATE_UNIT
// disable MSVC warning about extern template (it is valid C++11)
#pragma warning( disable : 4231)

	extern template float  Resample_Kernel(float src[4][4], CoordF uv, RescaleInfo type, FractalRescaleInfo fractalinfo);
	extern template colorF Resample_Kernel(colorF src[4][4], CoordF uv, RescaleInfo type, FractalRescaleInfo fractalinfo);
#endif

template< RESCALE_TYPE Method, typename T >
void RescaleExisting( WMGridPacket<T> const&src, WMGridPacket<T> &dest, RescaleInfo method )
{
	method.type = Method;
	CoordF scale;
	scale.x = float(src.w()-1.0f) / float(dest.w()-1.0f);
	scale.y = float(src.h()-1.0f) / float(dest.h()-1.0f);

	const int rmin = RescaleTypeTraits<Method>::zeroOffset;
	const int kmin = RescaleTypeTraits<Method>::kernelMin;
	const int kmax = RescaleTypeTraits<Method>::kernelMax;

	// Do the main loop across destination, expanding or reducing each dimension as we go
	for (int y = 0; y < dest.h(); y++) {
		for (int x = 0; x < dest.w(); x++) {
			CoordF inner_loc, uv;
			inner_loc.x = (scale.x * float(x));
			inner_loc.y = (scale.y * float(y));
			uv.x = inner_loc.x - floorf(inner_loc.x);
			uv.y = inner_loc.y - floorf(inner_loc.y);

			Coord  srcloc(int(inner_loc.x), int(inner_loc.y));
			T data[4][4];
			for (int j = kmin; j <= kmax; ++j) {
				for (int i = kmin; i <= kmax; ++i) {
					data[i][j] = src.valueSafe(Coord(srcloc.x+i-rmin, srcloc.y+j-rmin));
				}
			}
			dest[Coord(x,y)] = Resample_Kernel(data, uv, method);
		}
	}
}

} // end rescale
} // end wm
