#pragma once
// The following headers are required for all allocators.
#include <stddef.h> 
#include <new>      
#include <stdexcept> 

/** 
 Custom STL allocator that uses the World Machine memory functions and works across DLL boundaries.

 This is necessary because when we wish to pass a string around between the app and the device plugin DLLs but always have
 allocations and deletions happen from the app memory heap.

 Implementation is similar that shown in the "Mallocator" described here: http://blogs.msdn.com/b/vcblog/archive/2008/08/28/the-mallocator.aspx
 Most of the below is just boilerplate.. the important functions are the allocate and deallocate calls.
*/
template<class T>
class wmcore_allocator {
public:	
// typedefs required of all STL allocators
    typedef T * pointer;
    typedef const T * const_pointer;
    typedef T& reference;
    typedef const T& const_reference;
    typedef T value_type;
    typedef size_t size_type;
    typedef ptrdiff_t difference_type;

	// Empyty constructors and destructor
    wmcore_allocator() {}
    wmcore_allocator(const wmcore_allocator&) {}
    template <typename U> wmcore_allocator(const wmcore_allocator<U>&) {}
     ~wmcore_allocator() { }


    T * address(T& r) const {
        return &r;
    }
	const T * address(const T& s) const {
        return &s;
    }

	// The two ACTUAL functions of note here! The bodies are defined below the class defition.
 	pointer allocate(size_type n, const void*hint = 0);
	void deallocate(pointer p, size_type);

	// maximum possible size of an allocation.. in other words, the maximum unsigned value of size_t
    size_t max_size() const {
        return (static_cast<size_t>(0) - static_cast<size_t>(1)) / sizeof(T);
    }
 
    // The following must be the same for all allocators.
    template <typename U> struct rebind {
        typedef wmcore_allocator<U> other;
    };
 
    bool operator!=(const wmcore_allocator& other) const {
        return !(*this == other);
    }
 
	// in-place construction
    void construct(T * const p, const T& t) const {
        void * const pv = static_cast<void *>(p);
 
        new (pv) T(t);
    }

	// in-place destructor
    void destroy(T * const p) const;
 
 
    // Returns true if and only if storage allocated from *this
    // can be deallocated from other, and vice versa.
    // Always returns true for stateless allocators.
    bool operator==(const wmcore_allocator& other) const {
        return true;
    }
 
};

//////////////////////////////////////////////////////////////////////////
// Implementation
//////////////////////////////////////////////////////////////////////////



#include "Globals.h"
// We must define the allocator and include WMGlobal first before implementing the functions below


// The definition of destroy() must be the same for all allocators.
template <typename T> void wmcore_allocator<T>::destroy(T * const p) const {
    p->~T();
}

// Allocate using our core, but don't allow this allocation to trigger paging. Thus this allocator is to be used for
// internal/core allocations as opposed to data allocations. Data allocations may trigger paging.
template<class T>
T* wmcore_allocator<T>::allocate(size_t n, const void* = 0)
{
	if(T* p = static_cast<T*>(WorldMachine::core->memory->Alloc(sizeof(T) * n, false)) )
		return p;
	throw std::bad_alloc();
};

template<class T>
void wmcore_allocator<T>::deallocate(T* p, size_type)
{
	WorldMachine::core->memory->Free(p);
};

/// Specialized std containers to use the WM allocator, allowing them to pass their contents across a DLL boundary.
/// These reside in the global namespace for compatibility reasons with old code.
//@{
typedef std::basic_string<char, std::char_traits<char>, wmcore_allocator<char> > WMString;
typedef std::vector<Device*, wmcore_allocator<Device*> > WMDeviceVector;
//@}
