#ifndef _COMMONDEF_H
#define _COMMONDEF_H
#pragma once
/** @file common.h
 Common include file
 for global declarations, etc.
*/

//#pragma warning (disable : 4305)
#define WIN32_LEAN_AND_MEAN

#ifndef _CRT_SECURE_NO_DEPRECATE
#define _CRT_SECURE_NO_DEPRECATE
#endif

#define PI 3.141592653589793f

#include <assert.h>
#include <math.h>
#include <iostream>
#include <boost/smart_ptr.hpp>
#include "ForwardDef.h"

// This lets us control if functions which are not virtual in an inheritance sense, but need to be virtual for proper decoupling and calling from the v-table instead of directly.
// 
#define WMINTERFACE virtual

extern std::ofstream logfile;

void LOG_DEBUG(int channel, const char *text);	///< use for debugging purposes. Will direct to the debug console. The older messages are based upon this function:

void  LOG_WRITE(const char *a);
#define LOG_ERROR(a)		{ LOG_WRITE(a); }		
#define LOG_WARNING(a)		{ LOG_WRITE(a); }

#ifndef byte
typedef unsigned char byte;
#endif

#ifndef word
typedef unsigned short word;
#endif

#ifndef dword
typedef unsigned long dword;
#endif

#define USE_x64_COMPATIBLE
#include <intrin.h>

// Neat pow2 functions from flipcode .. wish I'd thoughta them
inline bool IsPow2(int num) {
	return (num && !(num & (num-1)));
}

// Returns next power of two, or num if num == pow2;
inline int NextPow2(int num)
{
    unsigned int RetVal = 1;

    if (IsPow2(num))
		return num;

#ifdef USE_x64_COMPATIBLE
	dword val = 0;
	_BitScanReverse(&val, num);
	++val;
	RetVal <<= val;
#else
    __asm    
    {    
        xor ecx, ecx
        bsr ecx, num
        inc ecx
        shl RetVal, cl
    }
#endif
    return (RetVal);
}

// Returns previous power of two, or num if num == pow2;
inline int PrevPow2(int num)
{
    unsigned int RetVal = 1;

#ifdef USE_x64_COMPATIBLE
	dword val = 0;
	_BitScanReverse(&val, num);
	RetVal <<= val;
#else
    __asm    
    {    
        xor ecx, ecx
        bsr ecx, num
        shl RetVal, cl
    }
#endif
    return (RetVal);
}

// returns the exponent of the greatest power of two equal-to or less than num
// ie GetPower(2) returns 1, GetPower(16) returns 4.
inline int GetPower(int num) {
	dword power = 0;
#ifdef USE_x64_COMPATIBLE
	if (_BitScanReverse(&power, num))
		return power;
	else
		return 0;
#else
	__asm {
        xor ecx, ecx
        bsr ecx, num
		mov power, ecx;
	}
	return power;
#endif
}

union INTORFLOAT {
	int i;
	float f;
};

// neat fast float-to-int function from Game Programming Gems 2
inline int ffi(float val) {
#ifdef USE_x64_COMPATIBLE
	return int(val);
#else
	INTORFLOAT	n, bias;
	n.f = val;
	bias.i = ((23 + 127) << 23) + (1<<22);
	n.f += bias.f;
	n.i -= bias.i;
	return n.i;
#endif
}

class SCOPED_CRITSEC {
public:
	SCOPED_CRITSEC(CRITICAL_SECTION &crit) 
	{ 
		c = &crit; 
#ifdef _DEBUG
		if (crit.LockCount > 0) {
			char buffer[256];
			sprintf(buffer, "Contention on Critical Section %p", c);
			LOG_WRITE(buffer);
		}
#endif
		EnterCriticalSection(c); 
	};
	~SCOPED_CRITSEC() {
		LeaveCriticalSection(c); 
	};
private:
	CRITICAL_SECTION *c;
};

class CONTROLLED_CRITSEC {
public:
	CONTROLLED_CRITSEC(CRITICAL_SECTION &crit) 
	{ 
		holding_critsec = false;
		c = &crit; 
		EnterCriticalSection(c); 
		holding_critsec = true;
	};
	~CONTROLLED_CRITSEC() {
		if (holding_critsec)
			LeaveCriticalSection(c); 
		holding_critsec = false;
	};

	void Lock() {
		if (!holding_critsec) {
			EnterCriticalSection(c);
			holding_critsec = true;
		}
	};

	void Unlock() {
		if (holding_critsec) {
			LeaveCriticalSection(c);
			holding_critsec = false;
		}
	}
private:
	CRITICAL_SECTION *c;
	bool holding_critsec;
};

/// Change angle in degrees to angle in radians
inline float degree2rad(float ang) {
	return float(ang * (PI / float(180.0)));
};

/// Radians to Degrees
inline float rad2degree(float rad) {
	return float(rad * (180.0 / PI));
};

inline byte _r(dword c) { dword *cptr = &c; byte *bptr = (byte*) cptr; return bptr[0];}; 
inline byte _g(dword c) { dword *cptr = &c; byte *bptr = (byte*) cptr; return bptr[1];}; 
inline byte _b(dword c) { dword *cptr = &c; byte *bptr = (byte*) cptr; return bptr[2];}; 
inline byte _a(dword c) { dword *cptr = &c; byte *bptr = (byte*) cptr; return bptr[3];};

/// Bezier spline interpolation
template <class T> T spline_bezier(T a, T b, T c, T d, float u) {
	float omu = 1.0f - u;
	float omu_square = omu * omu;
	float omu_cubed = omu_square * omu;
	float squared = u * u;
	float cubed = u * squared;
	return (a * (omu_cubed) +
		    b * (omu_square * 3 * u) +
			c * (omu * 3 * squared) +
			d * cubed);
};

/// Returns the first derivative of a Bezier spline 
template <class T> T spline_bezier_firstder(T a, T b, T c, T d, float u) {
	float omu = 1.0f - u;
	float omu_square = omu * omu;
	float omu_cubed = omu_square * omu;
	float squared = u * u;
	float cubed = u * squared;
	return (a * (-3.0f * omu_square) +
		    b * (9.0f * squared - 12.0f*u + 3) +
			c * (-9.0f * squared  + 6.0f*u) +
			d * 3.0f * squared);
};


/// Catmul-Rom spline interpolation
template <class T> T spline_catmull(T a, T b, T c, T d, float u) {
	float squared = u * u;
	float cubed = u * squared;
	return (a * (-0.5f * cubed + squared - 0.5f*u) +
		    b * (1.5f * cubed - 2.5f*squared + 1.0f) +
		    c * (-1.5f * cubed + 2.0f*squared + 0.5f*u) +
		    d * (0.5f*cubed - 0.5f * squared));
};

/// B-Spline spline interpolation
template <class T> T spline_bspline(T a, T b, T c, T d, float u) {
	const float onesixth = 1.0f / 6.0f;
	const float onethird = 1.0f / 3.0f;

	float omu = 1.0f - u;
	float omu_cube = omu * omu *omu;
	float squared = u * u;
	float cubed = u * squared;
	return (a * (onesixth * omu_cube) +
		    b * (0.5f * cubed - squared + 2.0f*onethird) +
		    c * (-0.5f * cubed + 0.5f*squared + 0.5f*u + onesixth) +
		    d * (onesixth*cubed));
};

/// returns the first derivative of a cubic B-Spline
template <class T> T spline_bspline_firstder(T a, T b, T c, T d, float u) {
	const float threetwo = 3.0f / 2.0f;

	float omu = 1.0f - u;
	float omu_square = omu * omu;
	float omu_cube = omu_square *omu;
	float squared = u * u;
	float cubed = u * squared;
	return (a * (-0.5f * omu_square) +
		    b * (threetwo * squared - 2.0f*u) +
		    c * (-threetwo * squared + u + 0.5f) +
		    d * (0.5f * squared));
};

/// Mitchell-Netravali parameterized spline
/// B=1, C=0 : B spline
/// B=0, C=0.5 : catmull-rom
/// B=0, C=? : parameterized alpha spline
template <class T> T spline_mitchell(T a, T b, T c, T d, float u, float B, float C) {
	float squared = u * u;
	float cubed = u * squared;

	const float onesixth = 1.0f / 6.0f;
	const float onethird = 1.0f / 3.0f;
	const float fivetwo = 5.0f / 2.0f;

	const float term1 = (-onesixth*B -C) * cubed + (0.5f*B + 2.0f*C) * squared + (-0.5f*B - C)*u + onesixth*B;
	const float term2 = (2.0f - 1.5f*B - C) * cubed + (-3.0f +2.0f*B + C)*squared + (1.0f - onethird*B);
	const float term3 = (-2.0f + 1.5f*B + C) * cubed + (3.0f - fivetwo*B - 2.0f*C)*squared + (0.5f*B + C)*u  + onesixth*B;
	const float term4 =  (onesixth*B + C) * cubed + -C*squared;

	return a * term1 + b * term2 + c * term3 + d * term4;
};

/// Linear interpolation
template <class T> T lerp(T a, T b, float u) {
	return T(a + ((b - a))*u);
};

template <class T> T extrapolate(const T rangemin, const T rangemax, const T value) {
	return (value-rangemin) / (rangemax - rangemin);
}

/// Cosine interpolation
template <class T> T cos_interp(T a, T b, float x ) {
	float ft = x * 3.1415927;
	float f = (1.0 - cos(ft)) * 0.5;
	return a*(1.0-f) + b*f;
};

/// Sine^2 nonlinear smoothing
inline float nonlinear_smooth(float val) {
	float partial = sin(val * 1.57079632679f);
	return partial*partial;
}

/// clamp func
template <class T> void clamp(T &value, const T min, const T max) {
	if (value < min)
		value = min;
	else if (value > max)
		value = max;
	return;
};

/// swap
template <class T> void swapbyval(T &a, T &b) {
	T temp(a);
	a = b;
	b = temp;
}

/// Convenience function to read non-char data to C++ streams
/// This is for use only with plain old data types, not types with embedded non-POD members
template < typename T >
void read(std::istream &in, T &result, size_t count = sizeof(T)) {
	in.read(reinterpret_cast<char*>(&result), count);
}

/// Convenience function to write out non-char data to C++ streams.
/// This is for use only with plain old data types, not types with embedded non-POD members
template < typename T >
void write(std::ostream &out, T const&val, size_t count = sizeof(T)) {
	out.write(reinterpret_cast<const char*>(&val), count);
}


#include "Geometry.h"

#endif